package dev.shadowsoffire.apotheosis.ench.table;

import net.minecraft.util.random.WeightedEntry;
import dev.shadowsoffire.apotheosis.ench.EnchantmentInfo;
import dev.shadowsoffire.apotheosis.ench.EnchModule;
import net.minecraftforge.registries.ForgeRegistries;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;
import net.minecraft.Util;
import net.minecraft.util.random.WeightedRandom;
import java.util.stream.Collector;
import java.util.stream.Collectors;
import net.minecraft.world.item.enchantment.EnchantmentHelper;
import com.google.common.collect.Lists;
import net.minecraft.world.item.enchantment.EnchantmentInstance;
import java.util.List;
import net.minecraft.world.item.enchantment.Enchantment;
import java.util.Set;
import net.minecraft.util.Mth;
import net.minecraft.world.item.ItemStack;
import net.minecraft.util.RandomSource;

public class RealEnchantmentHelper
{
    public static int getEnchantmentCost(final RandomSource rand, final int num, final float eterna, final ItemStack stack) {
        final int level = Math.round(eterna * 2.0f);
        if (num == 2) {
            return level;
        }
        final float lowBound = 0.6f - 0.4f * (1 - num);
        final float highBound = 0.8f - 0.4f * (1 - num);
        return Math.max(1, Math.round(level * Mth.m_216267_(rand, lowBound, highBound)));
    }
    
    public static List<EnchantmentInstance> selectEnchantment(final RandomSource rand, final ItemStack stack, int level, final float quanta, final float arcana, final float rectification, final boolean treasure, final Set<Enchantment> blacklist) {
        final List<EnchantmentInstance> chosenEnchants = (List<EnchantmentInstance>)Lists.newArrayList();
        final int enchantability = stack.getEnchantmentValue();
        final int srcLevel = level;
        if (enchantability > 0) {
            final float quantaFactor = 1.0f + getQuantaFactor(rand, quanta, rectification);
            level = Mth.m_14045_(Math.round(level * quantaFactor), 1, (int)(EnchantingStatRegistry.getAbsoluteMaxEterna() * 4.0f));
            final ApothEnchantmentMenu.Arcana arcanaVals = ApothEnchantmentMenu.Arcana.getForThreshold(arcana);
            final List<EnchantmentInstance> allEnchants = getAvailableEnchantmentResults(level, stack, treasure, blacklist);
            final Map<Enchantment, Integer> enchants = EnchantmentHelper.m_44831_(stack);
            allEnchants.removeIf(e -> enchants.containsKey(e.f_44947_));
            final List<ArcanaEnchantmentData> possibleEnchants = allEnchants.stream().map(d -> new ArcanaEnchantmentData(arcanaVals, d)).collect((Collector<? super Object, ?, List<ArcanaEnchantmentData>>)Collectors.toList());
            if (!possibleEnchants.isEmpty()) {
                chosenEnchants.add(WeightedRandom.m_216822_(rand, (List)possibleEnchants).get().data);
                removeIncompatible(possibleEnchants, (EnchantmentInstance)Util.m_137509_((List)chosenEnchants));
                if (arcana >= 25.0f && !possibleEnchants.isEmpty()) {
                    chosenEnchants.add(WeightedRandom.m_216822_(rand, (List)possibleEnchants).get().data);
                    removeIncompatible(possibleEnchants, (EnchantmentInstance)Util.m_137509_((List)chosenEnchants));
                }
                if (arcana >= 75.0f && !possibleEnchants.isEmpty()) {
                    chosenEnchants.add(WeightedRandom.m_216822_(rand, (List)possibleEnchants).get().data);
                }
                final int randomBound = 50;
                if (level > 45) {
                    level = (int)(srcLevel * 1.15f);
                }
                while (rand.m_188503_(randomBound) <= level) {
                    if (!chosenEnchants.isEmpty()) {
                        removeIncompatible(possibleEnchants, (EnchantmentInstance)Util.m_137509_((List)chosenEnchants));
                    }
                    if (possibleEnchants.isEmpty()) {
                        break;
                    }
                    chosenEnchants.add(WeightedRandom.m_216822_(rand, (List)possibleEnchants).get().data);
                    level /= 2;
                }
            }
        }
        return ((IEnchantableItem)stack.m_41720_()).selectEnchantments(chosenEnchants, rand, stack, srcLevel, quanta, arcana, treasure);
    }
    
    public static void removeIncompatible(final List<ArcanaEnchantmentData> list, final EnchantmentInstance data) {
        final Iterator<ArcanaEnchantmentData> iterator = list.iterator();
        while (iterator.hasNext()) {
            if (!data.f_44947_.m_44695_(iterator.next().data.f_44947_)) {
                iterator.remove();
            }
        }
    }
    
    public static List<EnchantmentInstance> getAvailableEnchantmentResults(final int power, final ItemStack stack, boolean allowTreasure, final Set<Enchantment> blacklist) {
        final List<EnchantmentInstance> list = new ArrayList<EnchantmentInstance>();
        final IEnchantableItem item = (IEnchantableItem)stack.m_41720_();
        allowTreasure = item.isTreasureAllowed(stack, allowTreasure);
        for (final Enchantment enchantment : ForgeRegistries.ENCHANTMENTS) {
            final EnchantmentInfo info = EnchModule.getEnchInfo(enchantment);
            if (!info.isTreasure() || allowTreasure) {
                if (!info.isDiscoverable()) {
                    continue;
                }
                if (blacklist.contains(enchantment)) {
                    continue;
                }
                if (!enchantment.canApplyAtEnchantingTable(stack) && !item.forciblyAllowsTableEnchantment(stack, enchantment)) {
                    continue;
                }
                for (int level = info.getMaxLevel(); level > enchantment.m_44702_() - 1; --level) {
                    if (power >= info.getMinPower(level) && power <= info.getMaxPower(level)) {
                        list.add(new EnchantmentInstance(enchantment, level));
                        break;
                    }
                }
            }
        }
        return list;
    }
    
    public static float getQuantaFactor(final RandomSource rand, final float quanta, final float rectification) {
        final float gaussian = (float)rand.m_188583_();
        float factor = Mth.m_14036_(gaussian / 3.0f, -1.0f, 1.0f);
        final float rectPercent = rectification / 100.0f;
        if (factor < rectPercent - 1.0f) {
            factor = Mth.m_216267_(rand, rectPercent - 1.0f, 1.0f);
        }
        return quanta * factor / 100.0f;
    }
    
    public static class ArcanaEnchantmentData extends WeightedEntry.IntrusiveBase
    {
        EnchantmentInstance data;
        
        public ArcanaEnchantmentData(final ApothEnchantmentMenu.Arcana arcana, final EnchantmentInstance data) {
            super(arcana.getRarities()[data.f_44947_.m_44699_().ordinal()]);
            this.data = data;
        }
    }
}
